/*
For information on this software or to report bugs please visit http://www.ranton.org
or email ranton.org

Copyright (c) 2002, Richard Neal Anton
All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

*	Redistributions of source code must retain the above copyright notice, this list
	of conditions and the following disclaimer. 

*	Redistributions in binary form must reproduce the above copyright notice, this list
	of conditions and the following disclaimer in the documentation and/or other materials
	provided with the distribution. 

*	The name of the copyright holder may not be used to endorse or promote products 
	derived from this software without specific prior written permission. 

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY
EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT
SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT
OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

// NOTE: I think this _always_ honors absolute paths.
// NOTE: getopt() argument parsing would be much better.
// NOTE: an interface script to deal with piping stuff in would be nice.

// I put in the magic number divider between file data,
// and did it in such a way that we can still read old archives.
// it doesn't write it for directories because they don't have
// any data associated with them anyway.

// Ok, it's now incompatible anyway because I changed it
// to not write the atime, we need to write a version
// number and file identification value at the start of
// the archive.

// Ok, we have a problem, we are hitting the 2GB limit of the
// ext2 file system like a brick wall, so we need to implement
// some kind of segmenting, preferably so that any individual segment
// is independent from the others, so that we have greater recoverability.
// Actually, that would be a bitch, because we can't segment
// to a specific size if we do it so that they're independent.

// It would also be good if we changed the incremental
// mode to store the attributes for all the files, not just the
// ones it stored the data for that changed so that we could
// support incremental backups past level 1.

// Another thing we should do is put in CRC calculation
// and storage, and store it locally with the 
// divider so when trashed archives are being picked apart
// it will be possible to tell which files are OK for sure.

// This now does everything essential,
// it could still really use some niceties
// though, like extracting specific files only,
// and better command line option setup
// to turn the force flag on and off and stuff
// like that

#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <utime.h>
#include <errno.h>
#include <assert.h>
#include <string.h>

// you may need to uncomment the below line for Linux
// #include <asm/types.h> // blatant linux dependency

// runtime flags for verbosity and debugging,
// will be controllable from options eventually
int verbose = 1, debug = 0;

// max path len not counting null
// this is what posix defines it to be
#define kMaxPathLen 255

// size of buffer used for catting files.
#define kCatBufSize 40960

// this is the written in the file size
// of an attr entry
#define kAttrSize (kMaxPathLen+1 + 3*2 + 4*4)

typedef unsigned short bu16;
typedef unsigned long bu32;

typedef struct file_attrs {
	char path[kMaxPathLen+1];
	bu16 mode, uid, gid;
	bu32 size, ctime, mtime;//, atime;
	bu32 pos; // offset of file data in archive file if regular file
//	int incFlag; // this flag is used during an incremental backup
		// to indicate that this file should be written into
		// the new archive.
	struct file_attrs *next; // for the temp list
} file_attrs;

// this is our error reporting function.
void report_err(int inErr, char *inDesc);

// this prints out a file_attrs structure
void print_attrs(file_attrs *inAttrs);

// this writes a 16 bit value little endian
int write_16b_le(bu16 inVal, int in_fd);
// this reads a 16 bit value little endian
int read_16b_le(bu16 *outVal, int in_fd);
// this writes a 32 bit value little endian
int write_32b_le(bu32 inVal, int in_fd);
// this reads a 32 bit value little endian
int read_32b_le(bu32 *outVal, int in_fd);

// this fills in our file_attrs record for
// a file with the given pathname
int get_file_attrs(char *inPath, file_attrs *outAttrs);

// this restores all of the desired file attrs
int set_file_attrs(file_attrs *inAttrs);

// this reads the attrs to the file
int read_attrs(file_attrs *outAttrs, int in_fd);

// this writes the attrs to the file
int write_attrs(file_attrs *inAttrs, int in_fd);

// this cats/appends file inAttrs->path onto the fd outOnto
int cat_file_onto(file_attrs *inAttrs, int outOnto);

// this writes a file from out of the archive
int cat_file_outof(file_attrs *inAttrs, int outOf);

// this is our qsort comparison function
int compare_attrs(const void *inA, const void *inB);

// decide if a file has changed.
int file_has_changed( file_attrs *inA, file_attrs *inB );

// this returns true if the path is a dir,
// 0 if it isn't, and -1 and sets errno
// if there was an error.
int path_is_dir( char *inPath );

// this returns true if the path exists
// false if it doesn't, and -1 and sets errno if there
// was an error other than ENOENT
int path_exists( char *inPath );

// this creates the array from the linked list.
file_attrs *create_array( file_attrs *head, long fcnt );

// this creates a file from the stored data and the
// file_attrs record.
int create_file( file_attrs *attrs, int fd, int forceFlag );

int write_magic_divider( int fd );
#define kMagicDividerSize 8

// these are our constants for mode
typedef enum {
	kCreating = 0,
	kExtracting,
	kIncrementing,
	kListing
} BarcMode;

int main(int argc, char **argv)
{
	char path[kMaxPathLen+2]; // have to account for newline and null
	char linkPath[kMaxPathLen+1];
	file_attrs attrs;
	int err = 0;
	int slen;
	int fd = -1;
//	int creating;
	BarcMode mode;
//	off_t cpos, 
	off_t written = 0;
	file_attrs *ptr, *head = NULL; // for the temp list of them,
		// before we sort them.
	file_attrs *old_array = NULL,
		*array = NULL; // for the array of them we
		// build and sort after we know how many there are.
	bu32 findex,old_cnt,fcnt = 0; // number of files
	int forceFlag = 1;

// this being on, means we are piping to stdout,
// and thus we _must_ turn off all debugging
// and verbosity output to stdout

	if( argc != 3 && argc != 4)
		goto usage;

	if( !strcmp( argv[1], "c" ) )
		mode = kCreating;
	else if( !strcmp( argv[1], "x" ) )
		mode = kExtracting;
	else if( !strcmp( argv[1], "i" ) )
		mode = kIncrementing;
	else if( !strcmp( argv[1], "l" ) )
		mode = kListing;
	else
		goto usage;
	
	if( mode == kIncrementing )
	{
		if( argc != 4 )
			goto usage;
	}
	else 
	{
		if( argc != 3 )
			goto usage;
	}

	if( mode == kCreating )
	{
		// this being on, means we are piping to stdout,
		// and thus we _must_ turn off all debugging
		// and verbosity output to stdout
		if( !strcmp( argv[2], "-" ) )
		{
			// pipe to stdout
			fd = 1; // stdout
			verbose = 0; debug = 0; // this will rape the world
			// in this mode
		}
		else
		{
			fd = open( argv[2], O_CREAT | O_WRONLY | O_TRUNC, 00644 );
			if( fd == -1 )
			{
				err = errno;
				goto cleanup;
			}
		}

		if( verbose )
			printf( "Creating archive\n" );

		while( fgets( path, kMaxPathLen+1, stdin ) )
		{
		//	printf("path = %s\n",path);
			slen = strlen(path);
			// kill the newline on the end
			if( path[slen-1] == '\n' )
				path[slen-1] = 0; 
			err = get_file_attrs(path,&attrs);
			if( err )
				report_err(err,path);
			else if( !S_ISREG(attrs.mode) && !S_ISDIR(attrs.mode) &&
				!S_ISLNK(attrs.mode) )
			{
				if( verbose )
					printf("Skipping %s\n",path); // skip devices,fifos,sockets..
			}
			else
			{
				if( verbose )
					printf("Archiving %s\n",path);
				if( debug )
					print_attrs(&attrs);

				// save the position in the file.
				// had to change this so it didn't explode
				// when piping stuff to standard out.
				attrs.pos = written;
				/*
				cpos = lseek( fd, 0, SEEK_CUR );
				if( cpos == (off_t)-1L )
					err = errno;
				else	
					attrs.pos = cpos;
				*/
		//		err = write_attrs( &attrs, fd );
				if(!err)
				{
					if( S_ISLNK( attrs.mode ) )
					{
						// NOTE: store the data returned from readlink
						// in the archive
						attrs.size = readlink( attrs.path, linkPath, 
							kMaxPathLen );
						if( attrs.size == -1 )
							err = errno;
						else
						{
							// write it onto the file.
							if( write( fd, linkPath, attrs.size ) 
								!= attrs.size )
							{
								err = errno;
							}
						}
					}
					else if( S_ISREG( attrs.mode ) )
						err = cat_file_onto(&attrs,fd);
					else if( S_ISDIR( attrs.mode ) )
						attrs.size = 0;
				}	
				// try not to bail
				if( err )
					report_err(err,path);
				// keep track of how much crap we've written 
				// so we can save the positions.
				written += attrs.size;
				// Now, write a magic number value
				// between files to make recovery of corrupted
				// or partial archives easier.
				if( !S_ISDIR( attrs.mode ) )
				{
					err = write_magic_divider( fd );
					if( err )
						goto cleanup;
					written += kMagicDividerSize;
				}

				// now put it in the list.
				ptr = (file_attrs *)malloc( sizeof( file_attrs ) );
				if( !ptr )
				{
					err = ENOMEM;
					goto cleanup;
				}
				memcpy((void *)ptr, (void *)&attrs, sizeof( file_attrs ) );
				// put it at the first of the list, cause it's easiest.
				ptr->next = head;
				head = ptr;
				// increment the count
				fcnt++;

			}
		}
		if( verbose )
			printf("%lu files and directories\n",fcnt);

		// now, take the list, and put it in an array
		// and run qsort on it.
		array = create_array( head, fcnt );
		if( !array )
		{
			err = ENOMEM;
			goto cleanup;
		}

		if( debug )
			printf("Freeing list of %lu attrs records\n",fcnt);

		// free the linked list.
		while(head)
		{
			ptr = head->next;
			free((void *)head);
			head = ptr;
		}

		if( debug )
			printf("Sorting %lu attrs records\n",fcnt);
		// sort the array
		qsort( (void *)array,fcnt, sizeof(file_attrs), compare_attrs );
		// now write out the atrributes onto the end of the file,
		// and then the count of them.
		if( debug )
		{
			long pos = lseek(fd,0,SEEK_CUR);
			printf("Writing %lu attrs records at %ld\n",fcnt,pos);

		}
		for( findex=0; findex<fcnt; findex++ )
		{
			if( debug )
				print_attrs( array+findex );
			err = write_attrs( array+findex, fd );
			if( err )
				goto cleanup;
		}
		err = write_32b_le( fcnt, fd );
		if( err )
			goto cleanup;
		// the array will get freed as soon as we hit cleanup,
		// so don't worry about it.
	}
	else if( mode == kExtracting )
	{
		if( verbose )
			printf( "Extracting archive\n");

		// extracting
		fd = open( argv[2], O_RDONLY );
		if( fd == -1 )
		{
			err = errno;
			goto cleanup;
		}
		// seek to the end of the file -4 and
		// read the number of files/dirs/whatevers

		if( lseek(fd,-4,SEEK_END) == -1 )
		{
			err = errno;
			goto cleanup;
		}
		// read the number of attrs entries.
		err = read_32b_le( &fcnt, fd );
		if( err )
			goto cleanup;
		// position to the start of
		// the array of attr entries.
		{
			long seekTo = -(4+kAttrSize*fcnt);
			if( debug )
				printf("reading attrs from %ld bytes before end\n",seekTo);
			if( lseek(fd,seekTo,SEEK_END) == -1 )
			{
				err = errno;
				goto cleanup;
			}
		}
		// allocate the array and read them in.
		array = (file_attrs *)malloc( sizeof(file_attrs) * fcnt );
		if( !array )
		{
			err = ENOMEM;
			goto cleanup;
		}
		for( findex=0;findex<fcnt;findex++ )
		{
			err = read_attrs( array+findex, fd );
			if( err )
				goto cleanup;
			if( debug )
				print_attrs( array+findex );
		}

		// now read in and write out the files
		for( findex = 0;findex<fcnt;findex++ )
		{
			// position to the correct place
			if( lseek( fd, array[findex].pos, SEEK_SET ) == -1 )
			{
				err = errno;
				goto cleanup;
			}

			// create the file, dir, or symlink
			err = create_file( array + findex, fd, forceFlag );
			
			if( err )
				report_err(err,array[findex].path);
				// try to keep going
			else
			{
				// set the file attrs.
				err = set_file_attrs(array+findex);
				if( err ) // report these errors but ignore them otherwise.
					report_err(err,array[findex].path);
			}
		}

	}
	else if( mode == kIncrementing )
	{
		// we are doing an incremental update,
		// need to read the attrs records from
		// the original archive, as well as create
		// a list of current ones.
		
		// so open the old archive first, before
		// we do anything else.
		fd = open( argv[2], O_RDONLY );
		if( fd == -1 )
		{
			err = errno;
			goto cleanup;
		}
		// seek to the end of the file -4 and
		// read the number of files/dirs/whatevers

		if( lseek(fd,-4,SEEK_END) == -1 )
		{
			err = errno;
			goto cleanup;
		}
		// read the number of attrs entries.
		err = read_32b_le( &old_cnt, fd );
		if( err )
			goto cleanup;
		// position to the start of
		// the array of attr entries.
		{
			long seekTo = -(4+kAttrSize*old_cnt);
			if( debug )
				printf("reading attrs from %ld bytes before end\n",seekTo);
			if( lseek(fd,seekTo,SEEK_END) == -1 )
			{
				err = errno;
				goto cleanup;
			}
		}
		// allocate the array and read them in.
		old_array = (file_attrs *)malloc( sizeof(file_attrs) * old_cnt );
		if( !old_array )
		{
			err = ENOMEM;
			goto cleanup;
		}
		for( findex=0;findex<old_cnt;findex++ )
		{
			err = read_attrs( old_array+findex, fd );
			if( err )
				goto cleanup;
			if( debug )
				print_attrs( old_array+findex );
		}
		// now close that file.
		if( close( fd ) )
		{
			err = errno;
			goto cleanup;
		}
		// we also need to allow the "-" specification
		// for the new archive as well.


		// NOW open the new archive.
		if( !strcmp( argv[3], "-" ) )
		{
			// pipe to stdout
			fd = 1; // stdout
			verbose = 0; debug = 0; // this will rape the world
			// in this mode
		}
		else
		{
			fd = open( argv[3], O_CREAT | O_WRONLY | O_TRUNC, 00644 );
			if( fd == -1 )
			{
				err = errno;
				goto cleanup;
			}
		}
		if( verbose )
			printf("Creating incremental archive\n");

		// now for each file we get to archive,
		// look for it in the old_array, and if it
		// isn't there or it has been modified,
		// then write it to our new archive.
		while( fgets( path, kMaxPathLen+1, stdin ) )
		{
		//	printf("path = %s\n",path);
			slen = strlen(path);
			// kill the newline on the end
			if( path[slen-1] == '\n' )
				path[slen-1] = 0; 
			err = get_file_attrs(path,&attrs);
			if( err )
				report_err(err,path);
			else if( !S_ISREG(attrs.mode) && !S_ISDIR(attrs.mode) &&
				!S_ISLNK(attrs.mode) )
			{
				if( verbose )
					printf("Skipping %s\n",path); // skip devices,fifos,sockets..
			}
			else
			{
				// see if we need this one in the new archive or not.
				ptr = (file_attrs *)bsearch( &attrs, (void *)old_array,
					old_cnt, sizeof( file_attrs ), compare_attrs );
		//		attrs.incFlag = 0;
				if( !ptr || file_has_changed( &attrs, ptr ) )
				{
		//			attrs.incFlag = 1;
					if( verbose )
						printf("Archiving %s\n",path);
					if( debug )
						print_attrs(&attrs);
	
					// save the position in the file.
					// NOTE: this makes it blow chunks
					// when piping output to another program.
					attrs.pos = written; // that should make it happy
					/*
					cpos = lseek( fd, 0, SEEK_CUR );
					if( cpos == (off_t)-1L )
						err = errno;
					else	
						attrs.pos = cpos;
					*/
			//		err = write_attrs( &attrs, fd );
					if(!err)
					{
						if( S_ISLNK( attrs.mode ) )
						{
							// NOTE: store the data returned from readlink
							// in the archive
							attrs.size = readlink( attrs.path, linkPath, 
								kMaxPathLen );
							if( attrs.size == -1 )
								err = errno;
							else
							{
								// write it onto the file.
								if( write( fd, linkPath, attrs.size ) 
									!= attrs.size )
								{
									err = errno;
								}
							}
						}
						else if( S_ISREG( attrs.mode ) )
							err = cat_file_onto(&attrs,fd);
						else if( S_ISDIR( attrs.mode ) )
							attrs.size = 0;
					}	
					// try not to bail
					if( err )
						report_err(err,path);
					// keep track of how much we've written so we
					// can save file position.
					written += attrs.size;

					// Now, write a magic number value
					// between files to make recovery of corrupted
					// or partial archives easier.
					if( !S_ISDIR( attrs.mode ) )
					{
						err = write_magic_divider( fd );
						if( err )
							goto cleanup;
						written += kMagicDividerSize;
					}

					// now put it in the list.
					ptr = (file_attrs *)malloc( sizeof( file_attrs ) );
					if( !ptr )
					{
						err = ENOMEM;
						goto cleanup;
					}
					memcpy((void *)ptr, (void *)&attrs, sizeof( file_attrs ) );
					// put it at the first of the list, cause it's easiest.
					ptr->next = head;
					head = ptr;
					// increment the count
					fcnt++;
	
				}
			}
		}
		// we're through with old_array now, so ditch it.
		free( (void *)old_array );
		old_array = NULL;

		if( verbose )
			printf("%lu files and directories\n",fcnt);
		// now we should have a count of the files we are
		// actuall adding, which are the only ones we
		// added to the list.

		// this creates the array from the linked list.
		array = create_array( head, fcnt );
		if( !array )
		{
			err = ENOMEM;
			goto cleanup;
		}
		
		if( debug )
			printf("Freeing list of %lu attrs records\n",fcnt);

		// free the linked list.
		while(head)
		{
			ptr = head->next;
			free((void *)head);
			head = ptr;
		}

		if( debug )
			printf("Sorting %lu attrs records\n",fcnt);
		// sort the array
		qsort( (void *)array,fcnt, sizeof(file_attrs), compare_attrs );
		// now write out the atrributes onto the end of the file,
		// and then the count of them.
		if( debug )
			printf("Writing %lu attrs records\n",fcnt);
		for( findex=0; findex<fcnt; findex++ )
		{
			if( debug )
				print_attrs( array+findex );
			err = write_attrs( array+findex, fd );
			if( err )
				goto cleanup;
		}
		err = write_32b_le( fcnt, fd );
		if( err )
			goto cleanup;
		// the array will get freed as soon as we hit cleanup,
		// so don't worry about it.
	}
	else if( mode == kListing )
	{
		// listing the archive.
		if( verbose )
			printf( "Listing archive\n");

		// extracting
		fd = open( argv[2], O_RDONLY );
		if( fd == -1 )
		{
			err = errno;
			goto cleanup;
		}
		// seek to the end of the file -4 and
		// read the number of files/dirs/whatevers

		if( lseek(fd,-4,SEEK_END) == -1 )
		{
			err = errno;
			goto cleanup;
		}
		// read the number of attrs entries.
		err = read_32b_le( &fcnt, fd );
		if( err )
			goto cleanup;
		if( verbose )
			printf( "Listing %lu files\n", fcnt);

		// have to clear the EOF flag on some systems
	//	(void)lseek(fd,0,SEEK_SET);

		// position to the start of
		// the array of attr entries.
		{
			long seekTo = -(4+kAttrSize*fcnt);
			if( debug )
				printf("reading attrs from %ld bytes before end\n",seekTo);
			if( lseek(fd,seekTo,SEEK_END) == -1 )
			{
				err = errno;
				goto cleanup;
			}
		}
		// allocate the array and read them in.
		array = (file_attrs *)malloc( sizeof(file_attrs) * fcnt );
		if( !array )
		{
			err = ENOMEM;
			goto cleanup;
		}
		
		for( findex=0;findex<fcnt;findex++ )
		{
			err = read_attrs( array+findex, fd );
			if( err )
				goto cleanup;
			if( debug )
				print_attrs( array+findex );
		}
		
		for( findex=0;findex<fcnt;findex++ )
		{
			// print out the file,
			printf("%s\n",array[findex].path);
			// NOTE: should print out ownership and permissions
			// too if it's verbose, but leaving it for later
		}
	}
	
	goto cleanup;
usage:
	fprintf(stderr,"Usage:\n");
	fprintf(stderr,"\tbarc c <archive>\n");
	fprintf(stderr,"\tbarc x <archive>\n");
	fprintf(stderr,"\tbarc i <old_archive> <incremental_archive>\n");
	fprintf(stderr,"\tbarc l <archive>\n");

cleanup:
	// if we have any memory allocated, then free it.
	if( head )
	{
		while(head)
		{
			ptr = head->next;
			free( (void *)head );
			head = ptr;
		}
	}
	if( array )
		free( (void *)array );

	// if we have a file descriptor that isn't stdout close it
	if( fd != -1 && fd != 1 )
		close( fd );
	if( err )
		report_err(err,"Error");
	else if( (err = ferror( stdin ) ) != 0 )
		report_err(err, "Couldn't read input");
	return err;
}

// decide if a file has changed since the last archived version.
int file_has_changed( file_attrs *inA, file_attrs *inB )
{
	// ok, if the size, the mtime, the atime, the ctime,
	// the uid, the gid, or the mode has changed then
	// we need to return true.
	if( 
		inA->mode != inB->mode ||
		inA->uid != inB->uid ||
		inA->gid != inB->gid ||
		inA->size != inB->size ||
// this is changed by reading, definately need to ignore it.
//		inA->atime != inB->atime ||
		inA->mtime != inB->mtime ||
		inA->ctime != inB->ctime )
			return 1;
	else
		return 0;
}

static char sCatBuf[kCatBufSize];
static int sCatBufInUse = 0;

// this cats/appends file inAttrs->path onto the fd outOnto
int cat_file_onto(file_attrs *inAttrs, int outOnto)
{
	int in_fd = -1;
	int err = 0;
	off_t chunk, left;

	assert(inAttrs);
	assert(outOnto >= 0 );


	// we use the flag so we can atleast tell if someone
	// is fucking calling us in a dangerous way.
	assert( !sCatBufInUse );
	sCatBufInUse = 1;
	in_fd = open(inAttrs->path,O_RDONLY);
	if( in_fd == -1 )
	{
		err = errno;
		goto cleanup;
	}
	// now read and write until done.
	left = inAttrs->size;
	if( debug )
		printf("size is %lu\n",(unsigned long)left);
	while( left > 0 )
	{
		chunk = kCatBufSize;
		if( chunk > left )
			chunk = left;
		if( debug )
			printf("chunk = %lu\n",(unsigned long)chunk);
		// read it in and write it back out.
		if( read(in_fd,sCatBuf,chunk) != chunk )
		{
			err = errno;
			assert( err && err != EAGAIN ); // want to know if we 
				// have to deal with this shit
			goto cleanup;
		}
		if( write( outOnto, sCatBuf, chunk) != chunk )
		{
			err = errno;
			assert( err && err != EAGAIN );
			goto cleanup;
		}
		left -= chunk;
	}

cleanup:
	sCatBufInUse = 0;
	if( in_fd != -1 )
		close( in_fd );
	return err;
}

// this writes a file from out of the archive
int cat_file_outof(file_attrs *inAttrs, int outOf)
{
	int out_fd = -1;
	int err = 0;
	off_t chunk, left;

	assert(inAttrs);
	assert(outOf >= 0 );


	// we use the flag so we can atleast tell if someone
	// is fucking calling us in a dangerous way.
	assert( !sCatBufInUse );
	sCatBufInUse = 1;
	out_fd = open(inAttrs->path,O_CREAT | O_WRONLY | O_TRUNC, 00644 );
	if( out_fd == -1 )
	{
		err = errno;
		goto cleanup;
	}
	// now read and write until done.
	left = inAttrs->size;
	if( debug )	
		printf("size is %lu\n",(unsigned long)left);
	while( left > 0 )
	{
		chunk = kCatBufSize;
		if( chunk > left )
			chunk = left;
		if( debug )
			printf("chunk = %lu\n",(unsigned long)chunk);
		// read it in and write it back out.
		if( read(outOf,sCatBuf,chunk) != chunk )
		{
			err = errno;
			assert( err && err != EAGAIN ); // want to know if we 
				// have to deal with this shit
			goto cleanup;
		}
		if( write( out_fd, sCatBuf, chunk) != chunk )
		{
			err = errno;
			assert( err && err != EAGAIN );
			goto cleanup;
		}
		left -= chunk;
	}

cleanup:
	sCatBufInUse = 0;
	if( out_fd != -1 )
		close( out_fd );
	return err;
}

// this fills in our file_attrs record for
// a file with the given pathname
int get_file_attrs(char *inPath, file_attrs *outAttrs)
{
	struct stat st_buf;

	assert( inPath );
	assert( outAttrs );
	assert( strlen(inPath) <= kMaxPathLen );

	// changed to call lstat
	if( lstat( inPath, &st_buf ) )
		return errno;

	// zero out the path field.
	memset( outAttrs->path, 0, kMaxPathLen+1 );
	strncpy( outAttrs->path, inPath, kMaxPathLen );
//	outAttrs->path[kMaxPathLen] = 0; // won't be null terminated
		// otherwise if strncpy hit max
	outAttrs->mode = st_buf.st_mode;
	outAttrs->uid = st_buf.st_uid;
	outAttrs->gid = st_buf.st_gid;
	outAttrs->size = st_buf.st_size;
	outAttrs->ctime = st_buf.st_ctime;
	outAttrs->mtime = st_buf.st_mtime;
//	outAttrs->atime = st_buf.st_atime;
	outAttrs->pos = (bu32)-1L; // flag it as not set yet

	return 0;
}

// this restores all of the desired file attrs
int set_file_attrs(file_attrs *inAttrs)
{
	struct utimbuf tbuf;
	int ret = 0;

	// NOTE: this function works once the file 
	// is already created.

	// set the owner and group
	if( chown( inAttrs->path, inAttrs->uid, inAttrs->gid ) )
	{
		ret = errno;
		// return error so the user gets a warning
		// but set the permissions properly before exiting.
	}

	// set the mode/permissions
	if( chmod( inAttrs->path, inAttrs->mode ) )
	{
		if( !ret )ret = errno; // keep earliest error
		goto cleanup;
	}

	// set the ctime/mtime/atime
//	tbuf.actime = inAttrs->atime;
	tbuf.modtime = inAttrs->mtime;
	if( utime( inAttrs->path, &tbuf ) )
	{
		if( !ret )ret = errno; // keep earliest error
		goto cleanup;
	}

cleanup:

	return ret;
}

// this writes the attrs to the file
int write_attrs(file_attrs *inAttrs, int in_fd)
{
	int err = 0;

	//slen = strlen( inAttrs->path );

	// write the path
	if( write( in_fd, inAttrs->path, kMaxPathLen+1 ) != kMaxPathLen+1 )
	{
		err = errno;
		assert( err && err != EAGAIN );
		goto cleanup;
	}
	// write the attributes
	err = write_16b_le(inAttrs->mode,in_fd);
	if( err )
		goto cleanup;
	err = write_16b_le(inAttrs->uid,in_fd);
	if( err )
		goto cleanup;
	err = write_16b_le(inAttrs->gid,in_fd);
	if( err )
		goto cleanup;
	err = write_32b_le(inAttrs->size,in_fd);
	if( err )
		goto cleanup;
	err = write_32b_le(inAttrs->ctime,in_fd);
	if( err )
		goto cleanup;
	err = write_32b_le(inAttrs->mtime,in_fd);
	if( err )
		goto cleanup;
//	err = write_32b_le(inAttrs->atime,in_fd);
//	if( err )
//		goto cleanup;

	err = write_32b_le(inAttrs->pos,in_fd);
	if( err )
		goto cleanup;

	
cleanup:
	return err;
}

// this reads the attrs to the file
int read_attrs(file_attrs *outAttrs, int in_fd)
{
	int err = 0;
	int ret;

	assert( outAttrs );

	// read the path
	if( (ret=read( in_fd, outAttrs->path, kMaxPathLen+1 )) != kMaxPathLen+1 )
	{
		if( debug )
			printf("Tried to read %d bytes, read() returned %d\n",kMaxPathLen+1,ret);

		report_err(errno,"Error reading file attributes");
		err = errno;
		goto cleanup;
	}
	// read the attributes
	err = read_16b_le(&(outAttrs->mode),in_fd);
	if( err )
		goto cleanup;
	err = read_16b_le(&(outAttrs->uid),in_fd);
	if( err )
		goto cleanup;
	err = read_16b_le(&(outAttrs->gid),in_fd);
	if( err )
		goto cleanup;
	err = read_32b_le(&(outAttrs->size),in_fd);
	if( err )
		goto cleanup;
	err = read_32b_le(&(outAttrs->ctime),in_fd);
	if( err )
		goto cleanup;
	err = read_32b_le(&(outAttrs->mtime),in_fd);
	if( err )
		goto cleanup;
//	err = read_32b_le(&(outAttrs->atime),in_fd);
//	if( err )
//		goto cleanup;

	err = read_32b_le(&(outAttrs->pos),in_fd);
	if( err )
		goto cleanup;

cleanup:
	return err;
}

// this writes a 16 bit value little endian
int write_16b_le(bu16 inVal, int in_fd)
{
	unsigned char ch[2];
	ch[0] = inVal&0xff;
	ch[1] = inVal>>8;
	if( write( in_fd, ch, 2 ) != 2 )
	{
		assert( errno && errno != EAGAIN );
		return errno;
	}
	return 0;
}

// this reads a 16 bit value little endian
int read_16b_le(bu16 *outVal, int in_fd)
{
	unsigned char ch[2];
	assert( outVal );
	if( read( in_fd, ch, 2 ) != 2 )
	{
		assert( errno && errno != EAGAIN );
		return errno;
	}
	*outVal = ch[1];
	*outVal <<= 8;
	*outVal |= ch[0];
	
	return 0;
}

// this writes a 32 bit value little endian
int write_32b_le(bu32 inVal, int in_fd)
{
	unsigned char ch[4];
	ch[0] = inVal&0xff;
	ch[1] = (inVal>>8)&0xff;
	ch[2] = (inVal>>16)&0xff;
	ch[3] = (inVal>>24)&0xff;
	if( write( in_fd, ch, 4 ) != 4 )
	{
		assert( errno && errno != EAGAIN );
		return errno;
	}
	return 0;
}

// this reads a 32 bit value little endian
int read_32b_le(bu32 *outVal, int in_fd)
{
	unsigned char ch[4];
	assert( outVal );
	
	if( read( in_fd, ch, 4 ) != 4 )
	{
		assert( errno && errno != EAGAIN );
		return errno;
	}
	*outVal = ch[3];
	*outVal <<= 8;
	*outVal |= ch[2];
	*outVal <<= 8;
	*outVal |= ch[1];
	*outVal <<= 8;
	*outVal |= ch[0];

	return 0;
}

// this prints out a file_attrs structure
void print_attrs(file_attrs *inAttrs)
{
	assert( inAttrs );
	printf(
	"%s \n\tuid %u gid %u mode %u\n", 
		inAttrs->path, inAttrs->uid, inAttrs->gid, inAttrs->mode );
	printf("\tctime %lu mtime %lu \n",
		inAttrs->ctime, inAttrs->mtime );
	printf("\tsize %lu pos %lu\n",
		inAttrs->size,  inAttrs->pos );
}

// this is our qsort comparison function
int compare_attrs(const void *inA, const void *inB)
{
	return strcmp( ((file_attrs *)inA)->path, 
		((file_attrs *)inB)->path );
}

// this returns true if the path is a dir,
// 0 if it isn't, and -1 and sets errno
// if there was an error.
int path_is_dir( char *inPath ) 
{
	struct stat st_buf;

	assert( inPath );

	if( stat( inPath, &st_buf ) )
		return -1;

	return S_ISDIR( st_buf.st_mode );
}

// this returns true if the path exists
// false if it doesn't, and -1 and sets errno if there
// was an error other than ENOENT
int path_exists( char *inPath )
{
	struct stat st_buf;
	int ret;

	assert( inPath );
	ret = stat( inPath, &st_buf );
	if( ret == 0 )
		return 1;
	else if( errno == ENOENT )
	{
		return 0;
		errno = 0; // to avoid confusing things
	}
	else
		return -1;
}

// this is our error reporting function.
void report_err(int inErr, char *inDesc)
{
	fprintf(stderr,"Error: %s %s(%d)\n",inDesc,strerror(inErr),inErr);
}

// this creates the array from the linked list.
file_attrs *create_array( file_attrs *head, long fcnt )
{
	file_attrs *array, *ptr;
	long findex;

	// now, take the list, and put it in an array
	// and run qsort on it.
	array = (file_attrs *)malloc( sizeof( file_attrs ) * fcnt );
	if( !array )
		return NULL;

	if( debug )
		printf("Moving list to array\n");
	ptr = head;
	findex = 0;
	while(ptr)
	{
		if( debug )
		{
			printf("record %lu of %lu\n",findex,fcnt);
			print_attrs( ptr );
		}
		memcpy( (void *)(array+findex), 
			(void *)ptr, sizeof( file_attrs ) );
		findex++;
		ptr = ptr->next;
	}
	assert( findex == fcnt );
	return array;
}

// this creates a file from the stored data and the
// file_attrs record.
int create_file( file_attrs *attrs, int fd, int forceFlag )
{
	int err = 0;
	char linkPath[kMaxPathLen+1];

	if( S_ISLNK(attrs->mode) )
	{
		// symlink
		// read in the symlink path
		if( read( fd, linkPath, attrs->size ) 
			!= attrs->size )
		{
			err = errno;
		}
		else
		{
			// terminate the string we read.
			linkPath[attrs->size] = 0;
			// NOTE: we need to check that it's
			// the same if it exists,
			// and write over it if it does.
			err = path_exists( attrs->path );
			if( err == 1 )
			{
				// unlink it.
				if( forceFlag )
				{
					if( unlink(attrs->path) )
						err = errno;
					else
						err = 0; 
				}
				else
					err = EEXIST;
			}
			else if( err )
			{
				err = errno;
			}
			// if we haven't fucked up yet go ahead.
			if( !err )
			{
				if( symlink( attrs->path, linkPath ) )
					err = errno;
			}
		}
	}
	else if( S_ISREG(attrs->mode) )
	{
		// if force is off, error if it's there already.
		if( path_exists( attrs->path ) )
		{
			// if we're not forcing, just error,
			if( !forceFlag )
				err = EEXIST;
			else
			{
				// if we are forcing, kill the directory,
				// or unlink the file or whatever.
				if( path_is_dir( attrs->path ) )
				{
					if( rmdir( attrs->path ) )
						err = errno;
				}
				else
				{
					// just go ahead and kill it so
					// it doesn't explode if the old one is a link
					if( unlink( attrs->path ) )
						err = errno;
				}
			}
		}
		if( !err )
			err = cat_file_outof( attrs, fd );
	}
	else if( S_ISDIR(attrs->mode) )
	{
		// mkdir, mode will get set later so don't worry about it.
		if( mkdir( attrs->path, 000 ) ) 
		{
			if(errno == EEXIST )
			{
				// it exists, as long as it's
				// a directory it's ok.
				if( !path_is_dir(attrs->path ) )
				{
					// it's not a directory, kill it, and
					// try again to create a new one.
					if( forceFlag )
					{
						if( unlink( attrs->path ) )
							err = errno;
						else
						{
							if( mkdir( attrs->path, 000 ) )
								err = errno;
						}
					}
					else
						err = errno;
				}
			}
			else
				err = errno;
		}	
	}


	return err;
}

int write_magic_divider( int fd )
{
	char *magicDiv = "*--Barc--*";// just something unlikely and 
		// easily recognizeable
	off_t ret;
	ret = write( fd, magicDiv, kMagicDividerSize );
	if( ret < 0 )
		return errno;
	else if( ret != kMagicDividerSize )
		return EIO;
	else
		return 0;
}


